'''
tables - support tables for user package under loutilities
==============================================================
'''
# standard
from urllib.parse import urlencode
from traceback import format_exception_only, format_exc

# pypi
from flask import g, current_app, url_for, jsonify
from flask.views import MethodView
from flask_security import auth_required
from flask_security import current_user
from sqlalchemy import Enum

# homegrown
from loutilities.tables import DbCrudApiRolePermissions, DteDbOptionsPickerBase
from loutilities.tables import SEPARATOR
from loutilities.user.model import Interest, db

class ParameterError(Exception): pass

debug = False

class DbCrudApiInterestsRolePermissions(DbCrudApiRolePermissions):
    '''
    This class extends DbCrudApiRolePermissions. This extension uses flask_security
    to do role checking for the current user and managing Interests usage (see
    loutilities.user
    '''

    decorators = [auth_required()]

    def __init__(self, **kwargs):
        if debug: current_app.logger.debug('DbCrudApiInterestsRolePermissions.__init__()')

        # the args dict has default values for arguments added by this derived class
        # caller supplied keyword args are used to update these
        # all arguments are made into attributes for self by the inherited class
        args = dict(
                    local_interest_model=None,
                    )
        args.update(kwargs)

        # this initialization needs to be done before checking any self.xxx attributes
        super().__init__(**args)

        # Caller must use local_interest_model
        if not self.local_interest_model:
            raise ParameterError('local_interest_model required')

    def permission(self):
        '''
        check for permission on data
        :rtype: boolean
        '''
        if debug: print('DbCrudApiInterestsRolePermissions.permission()')

        # need to be logged in
        if not current_user.is_authenticated:
            return False

        # g.interest initialized in <project>.create_app.pull_interest
        # g.interest contains slug, pull in interest db entry. If not found, no permission granted
        self.interest = Interest.query.filter_by(interest=g.interest).one_or_none()
        if not self.interest:
            return False

        # check permissions allowed/permitted
        if not super().permission():
            return False

        # current_user has permissions. Can this user access current interest?
        if self.interest in current_user.interests:
            return True
        else:
            return False

    def beforequery(self):
        '''
        filter on current interest
        :return:
        '''
        # self.interest set in self.permission()
        # need to convert to id of local interest table for query
        self.localinterest = self.local_interest_model.query.filter_by(interest_id=self.interest.id).one()
        self.queryparams['interest_id'] = self.localinterest.id

    def createrow(self, formdata):
        '''
        creates row in database

        :param formdata: data from create form
        :rtype: returned row for rendering, e.g., from DataTablesEditor.get_response_data()
        '''
        if debug: print('DbCrudApiInterestsRolePermissions.createrow()')

        # make sure we record the row's local interest id (self.localinterest set in self.beforequery())
        formdata['interest_id'] = self.localinterest.id

        # return the row
        row = super().createrow(formdata)

        return row

    def saformurl(self, **kwargs):
        '''
        standalone form url
        '''
        # NOTE: keyword arguments need to match request.args access in self.get()
        args = urlencode(kwargs)
        # self.__name__ is endpoint -- see https://github.com/pallets/flask/blob/master/flask/views.py View.as_view method
        url = '{}/saformjs?{}'.format(url_for('.' + self.my_view.__name__, interest=g.interest), args)
        return url

    def saformpostjs(self, saeditor):
        '''
        this gives subclass ability to add additional javascript code to saformjs handler
        (see self.get elif request.path[-9:] == '/saformjs':)

        this code is added after standalone form created, as saeditor

        :param saeditor: name of variable which holds standalone form
        :return: list of additional javascript strings
        '''
        js = super().saformpostjs(saeditor)

        js += [
            # note groups_groupname and groups_groupselectselector will now match that which was
            # used for register_group_for_editor
            '  // add event handlers for this standalone editor (see groups.js)',
            '  {}.groups_groupname = editor.groups_groupname;'.format(saeditor),
            '  {}.groups_groupselectselector = editor.groups_groupselectselector;'.format(saeditor),
            '  set_editor_event_handlers({});'.format(saeditor),
        ]

        return js

class AssociationSelect(DteDbOptionsPickerBase):
    '''
    AssociationSelect builds on DteDbOptionsPickerBase, allowing use of
    https://docs.sqlalchemy.org/en/13/orm/basic_relationships.html#association-object to
    add an Enum type selector for the relationship

    Additional parameters

    :param associationmodel: model class which contains [association object](https://docs.sqlalchemy.org/en/13/orm/basic_relationships.html#association-object)
    :param associationfields: list of fields in associationmodel used to update the record, must be Enum or subrecord
    :param selectattrs: list of attributes used to build select grouping. list order must match associationfields
            items must be Enum attribute (Enum) (uses Enum values) or table attribute (uses table values)

    Use of associationattr:

        # TODO: we'd like the select tree to look like this, but currently Editor does not support select2 with optgroup tag
        # select tree looks like
        #
        #     associationattr 1
        #         fieldmodel.id 1
        #         fieldmodel.id 2
        #             :
        #     associationattr 2
        #         fieldmodel.id 1
        #         fieldmodel.id 2
        #             :

        # associate task / taskfield tables adding association attribute "need"
        class TaskTaskField(Base):
            __tablename__ = 'task_taskfield'
            task_id             = Column(Integer, ForeignKey('task.id'), primary_key=True)
            taskfield_id        = Column(Integer, ForeignKey('taskfield.id'), primary_key=True)
            need                = Column(Enum('required', 'oneof', 'optional'))
            task                = relationship('Task', backref='fields')
            taskfield           = relationship('TaskField', backref='tasks')

        class Task(Base):
            __tablename__ = 'task'
            id                  = Column(Integer(), primary_key=True)
            interest_id         = Column(Integer, ForeignKey('localinterest.id'))
            interest            = relationship('LocalInterest', backref=backref('tasks'))
            task                = Column(String(TASK_LEN))

        class TaskField(Base):
            __tablename__ = 'taskfield'
            id                  = Column(Integer(), primary_key=True)
            interest_id         = Column(Integer, ForeignKey('localinterest.id'))
            interest            = relationship('LocalInterest', backref=backref('taskfields'))
            taskfield           = Column(String(TASKFIELD_LEN))
            fieldname           = Column(String(TASKFIELDNAME_LEN))

        taskfields = AssociationSelect(tablemodel=Task, fieldmodel=User, labelfield='name', formfield='users', dbfield='users', associationattr=TaskTaskField.need )
                        OR
        in configuration for DbCrudApiInterestsRolePermissions
            {'data': 'fields', 'name': 'fields', 'label': 'Fields',
             '_treatment': {
                 'relationship': {
                     'optionspicker':
                         AssociationSelect(
                             tablemodel=Task,
                             associationmodel=TaskTaskField,
                             associationtablemodelfield='task',
                             associationfields=['need', 'taskfield'],
                             selectattrs=[TaskTaskField.need, TaskField.taskfield],
                             labelfield='fields',
                             formfield='fields',
                             dbfield='fields', uselist=True,
                             queryparams=localinterest_query_params,
                         )
                 }}
             },


    '''
    def __init__(self, **kwargs):
        # the args dict has default values for arguments added by this class
        # caller supplied keyword args are used to update these
        # all arguments are made into attributes for self by the inherited class
        args = dict(
            associationmodel=None,
            associationtablemodelfield=None,
            associationfields=[],
            selectattrs=[],
        )
        args.update(kwargs)

        # associationmodel = TaskTaskField,
        # associationfields = ['need', 'taskfield'],
        # selectattrs = [TaskTaskField.need, TaskField.taskfield],

        # this initialization needs to be done before checking any self.xxx attributes
        super().__init__(**args)

        # some of the args are required
        reqdfields = ['associationmodel', 'associationtablemodelfield', 'associationfields', 'selectattrs', 'dbfield']
        for field in reqdfields:
            if not getattr(self, field, None):
                raise ParameterError('{} parameters are all required'.format(', '.join(reqdfields)))
        if len(self.associationfields) != len(self.selectattrs):
            raise ParameterError('length of associationfields and selectattrs must be the same')

        # pick up options for the tree levels
        self.optionlevels = []
        for associationfield, selectattr in zip(self.associationfields, self.selectattrs):
            thislevel = {'associationfield':associationfield, 'selectattr': selectattr}
            if type(selectattr.type) == Enum:
                thislevel['options'] = selectattr.type.enums
                thislevel['ids'] = range(len(selectattr.type.enums))
            else:
                thislevel['model'] = selectattr.class_
                thislevel['key'] = selectattr.key

            self.optionlevels.append(thislevel)

    # def _setid(self, tablemodelitem, id):
    #     theassociation = self.associationmodel(**{self.associationtablemodelfield:tablemodelitem})
    def _setid(self, id):
        if not id: return None

        theassociation = self.associationmodel()
        db.session.add(theassociation)

        idparts = id.split('.')
        for optionlevel,idpart in zip(self.optionlevels,idparts):
            # have to look up the target if this level is a model
            if 'model' in optionlevel:
                thetarget = optionlevel['model'].query.filter_by(id=idpart).one()
            else:
                thetarget = optionlevel['options'][int(idpart)]
            setattr(theassociation, optionlevel['associationfield'], thetarget)
        return theassociation

    def set(self, formrow):
        if self.uselist:
            # accumulate list of database model instances
            items = []

            # return empty list if no items, rather than list with empty item
            # this allows for multiple keys in formrow[self.formfield], but seems like there'd only be one
            itemvalues = []
            for key in formrow[self.formfield]:
                vallist = formrow[self.formfield][key].split(SEPARATOR)
                # empty list is actually null list with one entry
                if len(vallist) == 1 and not vallist[0]: continue
                # loop through nonempty entries -- will we ever see null entry? hope not else exception on .one() call below
                for ndx in range(len(vallist)):
                    if len(itemvalues) < ndx + 1:
                        itemvalues.append(vallist[ndx])
                    else:
                        itemvalues[ndx].update(vallist[ndx])
            if debug: current_app.logger.debug('itemvalues={}'.format(itemvalues))
            for itemvalue in itemvalues:
                thisitem = self._setid(itemvalue)
                items.append(thisitem)
            return items
        else:
            itemvalue = formrow[self.formfield] if formrow[self.formfield] else None
            thisitem = self._setid(itemvalue)
            return thisitem

    def _getlabelvalue(self, theitem):
        thislabel = ''
        thisvalue = ''
        for optionlevel in self.optionlevels:
            # startup case, don't append separators
            if thislabel != '':
                thislabel += '/'
                thisvalue += '.'

            # check if retrieving from the database, otherwise use existing options
            if 'model' in optionlevel:
                # targetid = getattr(theitem, optionlevel['associationfield'])
                # thetarget = optionlevel['model'].query.filter_by(id=targetid).one()
                thetarget = getattr(theitem, optionlevel['associationfield'])
                thislabel += getattr(thetarget, optionlevel['key'])
                thisvalue += str(getattr(thetarget, 'id'))
            else:
                theoptions = optionlevel['options']
                thislabel += getattr(theitem, optionlevel['associationfield'])
                thisvalue += str(theoptions.index(thislabel))

        return thislabel, thisvalue

    def get(self, dbrow_or_id):
        if type(dbrow_or_id) in [int, str]:
            dbrow = self.tablemodel.query().filter_by(id=dbrow_or_id).one()

        else:
            dbrow = dbrow_or_id

        # get from database to form
        if self.uselist:
            items = {}
            labelitems = []
            valueitems = []
            for item in getattr(dbrow, self.dbfield):
                thislabel,thisvalue = self._getlabelvalue(item)

                labelitems.append(thislabel)
                valueitems.append(thisvalue)

            items = {self.labelfield: SEPARATOR.join(labelitems), self.valuefield: SEPARATOR.join(valueitems)}
            return items
        else:
            # get the attribute if specified
            theitem = getattr(dbrow, self.dbfield)
            if theitem:
                thislabel, thisvalue = self._getlabelvalue(theitem)
                item = {self.labelfield: thislabel, self.valuefield: thisvalue}
                return item

            # otherwise return None
            else:
                return {self.labelfield: None, self.valuefield: None}

    def options(self):
        '''
        return sorted list of items in the model, may be overridden for more complex models

        :return: options as expected by optionpicker type,
            e.g., for select2 list of {'label': label, 'value': value} (see https://select2.org/options)
        '''
        queryparams = self.queryparams() if callable(self.queryparams) else self.queryparams
        items = [{'label':'', 'value':''}]

        for optionlevel in self.optionlevels:
            # overwrite options if retrieving from the database, otherwise use existing options
            if 'model' in optionlevel:
                optionsids = [(getattr(r,optionlevel['key']),getattr(r,'id'))
                              for r in optionlevel['model'].query
                                  .filter_by(**queryparams)
                                  .order_by(optionlevel['selectattr'])
                                  .all()]
                optionlevel['options'] = [oi[0] for oi in optionsids]
                optionlevel['ids'] = [oi[1] for oi in optionsids]

            # distribute these options across existing items
            theseitems = []
            for item in items:
                thislabel = item['label']
                thisvalue = item['value']
                # startup case, don't append separators
                if thislabel != '':
                    thislabel += '/'
                    thisvalue += '.'
                for option,id_ in zip(optionlevel['options'],optionlevel['ids']):
                    theseitems.append({'label':thislabel+option, 'value':thisvalue+str(id_)})
            items = theseitems

        if self.nullable:
            items =[{'label': '<none>', 'value': None}] + items

        return items

    def new_plus_options(self):
        '''
        return sorted list of items in the model, with first option being <new>

        :return:
        '''
        items = [{'label': '<new>', 'value': 0}] + self.options()
        return items

    def col_options(self):
        '''
        return additional column options required by the caller

        :return:
        '''
        col = {}
        col['type'] = 'select2'
        col['onFocus'] = 'focus'
        col['opts'] = {'minimumResultsForSearch': 0 if self.searchbox else 'Infinity',
                       'multiple': self.uselist,
                       'placeholder': None if self.uselist else '(select)'}
        if self.uselist:
            col['separator'] = SEPARATOR
        return col

class AssociationCrudApi(DbCrudApiInterestsRolePermissions):
    '''
    AssociationCrudApi MUST be used with AssociationSelect. This allows DbCrudApi... to update
    the association with the self.model instance, which isn't visible to AssociationSelect

    Additional parameters

    :param assnmodelfield: field in Association definition which points to the self.model dbrow being edited
    :param assnlistfield: field in self.model which has association record list or association record
    '''
    def __init__(self, **kwargs):

        # the args dict has default values for arguments added by this derived class
        # caller supplied keyword args are used to update these
        # all arguments are made into attributes for self by the inherited class
        args = dict(
            assnlistfield=None,
            assnmodelfield=None,
        )
        args.update(kwargs)

        # this initialization needs to be done before checking any self.xxx attributes
        super().__init__(**args)

        # Caller should use roles_accepted OR roles_required but not both
        reqdfields = ['assnlistfield', 'assnmodelfield']
        for field in reqdfields:
            if not getattr(self, field, None):
                raise ParameterError('{} are all required'.format(reqdfields))

    def updaterow(self, thisid, formdata):
        '''
        update assnlistfield in self.model, and assnmodelfield in AssociationSelect.associationmodel
        '''
        dbrow = self.model.query.filter_by(id=thisid).one()

        # get handy access to the association list field
        assnfield = getattr(dbrow, self.assnlistfield)

        # first delete all the items in assnlistfield
        for i in range(len(assnfield)):
            assnrow = assnfield.pop(0)
            db.session.delete(assnrow)

        # this adds the current fields list
        notused = super().updaterow(thisid, formdata)

        # now need to add this task to the tasktaskfield associations
        for assnrow in assnfield:
            setattr(assnrow, self.assnmodelfield, dbrow)

        return self.dte.get_response_data(dbrow)


class DbPermissionsMethodViewApi(MethodView):
    """
    method view which checks permissions, for apis

    :param app: app or blueprint this view belongs to
    :param db: database object a la sqlalchemy
    :param roles_accepted: list of roles accepted for this view
    :param endpoint: endpoint parameter used by flask.url_for()
    :param rule: rule parameter used by flask.add_url_rule()
    :param methods: sequence of HTTP methods the url rule applies to
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        args = dict(
            app=None,
            db=None,
            roles_accepted=None,
            endpoint=None,
            rule=None,
            methods=None,
        )
        args.update(kwargs)

        # make arguments into attributes
        for key in args:
            setattr(self, key, args[key])

    def permission(self):
        '''
        determine if current user is permitted to use the view
        '''
        allowed = False

        # g.interest initialized in <project>.create_app.pull_interest
        # g.interest contains slug, pull in interest db entry. If not found, no permission granted
        interest = Interest.query.filter_by(interest=g.interest).one_or_none()

        # need to be logged in, and allowed to use this interest
        if current_user.is_authenticated and interest and interest in current_user.interests:
            # check permissions allowed/permitted
            for role in self.roles_accepted:
                if current_user.has_role(role):
                    allowed = True
                    break

        return allowed

    def get(self):
        try:
            if not self.permission():
                db.session.rollback()
                cause = 'operation not permitted for user'
                return jsonify(error=cause)

            response = self.do_get()
            db.session.commit()
            return response

        except Exception as e:
            exc = ''.join(format_exception_only(type(e), e))
            output_result = {'status': 'fail', 'error': 'exception occurred:<br>{}'.format(exc)}
            # roll back database updates and close transaction
            db.session.rollback()
            current_app.logger.error(format_exc())
            return jsonify(output_result)

    def do_get(self):
        raise NotImplementedError

    def post(self):
        try:
            if not self.permission():
                db.session.rollback()
                cause = 'operation not permitted for user'
                return jsonify(error=cause)

            response = self.do_post()
            db.session.commit()
            return response

        except Exception as e:
            exc = ''.join(format_exception_only(type(e), e))
            output_result = {'status': 'fail', 'error': 'exception occurred:<br>{}'.format(exc)}
            # roll back database updates and close transaction
            db.session.rollback()
            current_app.logger.error(format_exc())
            return jsonify(output_result)

    def do_post(self):
        raise NotImplementedError

    def register(self):
        # name for view is last bit of fully named endpoint
        name = self.endpoint.split('.')[-1]

        self.my_view = self.as_view(name, **self.kwargs)
        self.app.add_url_rule(self.rule, view_func=self.my_view, methods=self.methods)




###########################################################################################
# blueprints - blueprint helpers
#
#       Date            Author          Reason
#       ----            ------          ------
#       07/21/18        Lou King        Create
#                                       adapted from https://gist.github.com/mattupstate/9257466
#
#   Copyright 2018 Lou King
#
###########################################################################################
'''
blueprints - blueprint helpers
=================================
'''

#------------------------------------------------------------------
def add_url_rules(bp, cls, decorator=None, decorator_args=[]):
#------------------------------------------------------------------
    '''
    add url rules to bp for class cls

    cls may define the following class attribute
        url_rules   dict {endpoint: options, ...} 
            endpoint    the endpoint for the registered URL rule
            options     tuple (url_rule[, methods[, defaults]])
                url_rule    the URL rule as string
                methods     tuple of supported methods, e.g. 'GET', 'POST'
                defaults    optional dict with defaults for other rules with the same endpoint
                            see http://werkzeug.pocoo.org/docs/0.14/routing/#werkzeug.routing.Rule
    '''
    for endpoint, options in cls.url_rules.items():
        url_rule = options
        methods = ('GET',)
        defaults = {}
        if len(options) == 2:
            url_rule, methods = options
        elif len(options) == 3:
            url_rule, methods, defaults = options

        # decorator may be specified. 
        # see http://flask.pocoo.org/docs/0.12/views/#decorating-views,
        # http://scottlobdell.me/2015/04/decorators-arguments-python/
        if not decorator:
            view_func = cls.as_view(endpoint)
        else:
            view_func = decorator(*decorator_args)(cls.as_view(endpoint))
        
        # add url rule
        bp.add_url_rule(url_rule, endpoint=endpoint, methods=methods,
                        defaults=defaults, view_func=view_func)

#------------------------------------------------------------------
def list_routes(app):
#------------------------------------------------------------------
    '''
    debug to list routes for app
    '''
    # adapted from http://flask.pocoo.org/snippets/117/
    from urllib import parse
    from flask import url_for
    output = []
    for rule in app.url_map.iter_rules():

        options = {}
        for arg in rule.arguments:
            options[arg] = "[{0}]".format(arg)

        methods = ','.join(rule.methods)
        url = url_for(rule.endpoint, **options)
        line = parse.unquote("{:50s} {:20s} {}".format(rule.endpoint, methods, url))
        output.append(line)
    
    for line in sorted(output):
        print(line)




'''
filters - filters which work with tables-assets/static/filters.css

Usage::

    filters = filtercontainerdiv()
    filters += filterdiv('filter1-id', 'Filter 1 Label')
    filters += filterdiv('filter2-id', 'Filter 2 Label')

        :

    yadcf_options = [
        yadcfoption(colnum, 'filter1-id', 'multi_select', uselist=True, placeholder='Select filter 1')
    ]

    example = CrudApi(
            :
        pretablehtml = filters.render(),
            :
        yadcfoptions = yadcf_options,
    )
'''

from dominate.tags import div, span

class ParameterError(Exception):
    '''
    raised if unknown or bad parameters seen
    '''

def filterdiv(id, label):
    '''
    build div with spans for label, filter

    :param id: html id of filter container, specified in filterdiv()
    :param label: label to display to the left of the filter
    :return: dominate div
    '''
    return div(
        span(label, _class='label'),
        span(id=id, _class='filter'),
        _class='filter-item',
    )

def filtercontainerdiv(**kwargs):
    '''
    build div for filter container

    :param kwargs: optional div() keyword arguments
    :return: dominate div
    '''
    return div(_class='external-filter filter-container', **kwargs)

def yadcfoption(colselector, filterid, filtertype, placeholder=None, width=None, date_format='yyyy-mm-dd',
                filter_match_mode='contains', **kwargs):
    '''
    return yadcf option entry for filter

    :param colselector: datatables column selector (see https://datatables.net/reference/type/column-selector)
    :param filterid: html id of filter container, specified in filterdiv()
    :param filtertype: see https://github.com/vedmack/yadcf/blob/master/src/jquery.dataTables.yadcf.js
    :param placeholder: (for select2) placeholder for empty filter
    :param width: (for select2) css value for width of filter container
    :param date_format: (for range_date) format for date, default 'yyyy-mm-dd'
    :param filter_match_mode: one of 'contains', 'exact', 'startsWith', default 'contains'
    :param kwargs: other keyword arguments for yadcf column options
    :return: yadcf option for inclusion in yadcf option array
    '''
    option = {
        'column_selector': colselector,
        'filter_container_id': filterid,
        'filter_type': filtertype,
        'filter_match_mode': filter_match_mode,
        'date_format': date_format,
    }
    option.update(**kwargs)

    if filtertype in ['select', 'multi_select']:
        option['select_type'] = 'select2'
        option['filter_reset_button_text'] = False

        option['select_type_options'] = {
            'allowClear': True,  # show 'x' (remove) next to selection inside the select itself
        }

        if placeholder:
            option['filter_default_label'] = placeholder
            # this is needed to set the id for filter Storage areas
            option['select_type_options']['placeholder'] = {
                'id': -1,
                'text': placeholder
            }

        if width:
            option['select_type_options']['width'] = width

        if filtertype == 'multi_select':
            option['text_data_delimiter'] = ', ',

    return option


#!/usr/bin/python
###########################################################################################
#   csvu - csv and string utilities
#
#   Date        Author      Reason
#   ----        ------      ------
#   11/25/13    Lou King    Create
#
#   Copyright 2013 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################

# standard
import unicodedata
import csv

# pypi
#from IPython.core.debugger import Tracer; debug_here = Tracer()

# github

# other

# home grown

#----------------------------------------------------------------------
def unicode2ascii(ustr):
#----------------------------------------------------------------------
    '''
    convert non-ascii unicode characters to ascii
    
    :param ustr: unicode or str
    :rtype: str
    '''
    if isinstance(ustr, str):
        return ustr
    else:
        return unicodedata.normalize('NFKD',ustr).encode('ascii','ignore')

#----------------------------------------------------------------------
def str2num(ustr):
#----------------------------------------------------------------------
    '''
    convert string to float, number, ascii
    
    :param ustr: unicode or str
    :rtype: int, float or str as appropriate, or None if ustr was None
    '''
    if ustr is None:
        return None
    
    try:
        return int(ustr)
    except ValueError:
        try:
            return float(ustr)
        except ValueError:
            return unicode2ascii(ustr).strip()

#######################################################################
class DictReaderStr2Num(csv.DictReader):
#######################################################################
    '''
    extend csv.DictReader to convert strings to numbers 
    '''

    #----------------------------------------------------------------------
    def __next__(self):
    #----------------------------------------------------------------------
        row = csv.DictReader.__next__(self)
        for key in row:
            row[key] = str2num(row[key])
        return row
        




#!/usr/bin/python
###########################################################################################
#   apikey - manage api keys for a given package
#
#   Date        Author      Reason
#   ----        ------      ------
#   04/01/13    Lou King    Create
#
#   Copyright 2013 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
apikey - manage api keys for a given package
===================================================
'''

# standard
import argparse

# pypi
import appdirs
# home grown
from . import version
from . import extconfigparser


# github
# other

class unknownKey(Exception): pass

########################################################################
class ApiKey():
########################################################################
    '''
    base class for API key management
    result of object creation is a apikeys.cfg file with an [apikeys] section, possibly empty
    
    keys are stored in the following locations
    * Windows: C:\\Users\\<username>\\AppData\\Local\\<author>\\<appname>\\apikeys.cfg
    * Linux: /home/<username>/.config/<appname>/apikeys.cfg
    * Mac OS X: /Users/<username>/Library/Application Support/<appname>/apikeys.cfg
    
    use :meth:`addkey` to add a key, :meth:`getkey` to retrieve a key
    
    :param author: author of software for which keys are being managed
    :param appname: application name for which keys are being managed
    '''
    
    keyssection = 'apikeys'
    
    #----------------------------------------------------------------------
    def __init__(self,author,appname):
    #----------------------------------------------------------------------
        '''
        '''

        configdir = appdirs.user_data_dir(appname,author)
        configfname = 'apikeys.cfg'
        self.cf = extconfigparser.ConfigFile(configdir,configfname)
        
    #----------------------------------------------------------------------
    def getkey(self,keyname):
    #----------------------------------------------------------------------
        '''
        return the key indicated by keyname
        if key doesn't exist, unknownKey is raised
        
        :param keyname: name of key for later retrieval
        :rtype: value of key
        '''
        try:
            return self.cf.get(self.keyssection,keyname)
        
        except extconfigparser.unknownSection:
            raise unknownKey
        
        except extconfigparser.unknownOption:
            raise unknownKey
        
    #----------------------------------------------------------------------
    def updatekey(self,keyname,keyvalue):
    #----------------------------------------------------------------------
        '''
        update or add a key to key file
        
        :param keyname: name of key for later retrieval
        :param keyvalue: value of key
        '''
        
        # write all the keys to a temporary file
        self.cf.update(self.keyssection,keyname,keyvalue)
        
    #----------------------------------------------------------------------
    def list(self):
    #----------------------------------------------------------------------
        '''
        Return a list of (keyname, value) pairs for each option in the given section.
        
        :rtype: [(keyname,value),...]
        '''
        return self.cf.items(self.keyssection)
        
#----------------------------------------------------------------------
def main():
#----------------------------------------------------------------------
    '''
    add key to key file
    '''

    parser = argparse.ArgumentParser(version='{0} {1}'.format('loutilities',version.__version__))
    parser.add_argument('application',help='name of application for which keys are to be stored')
    parser.add_argument('keyname',help='name of key to create/update in key configuration file',nargs='?',default=None)
    parser.add_argument('keyvalue',help='value of key to be put into key configuration file',nargs='?',default=None)
    parser.add_argument('-a','--author',help='name of software author. (default %(default)s)',default='Lou King')
    parser.add_argument('-l','--list',action='store_true',help='print list of keyname,values.  If set, keyname, value arguments are ignored')
    args = parser.parse_args()
    
    # act on arguments
    application = args.application
    author = args.author
    keyname = args.keyname
    keyvalue = args.keyvalue
    plist = args.list
    
    if not plist and not(keyname and keyvalue):
        print('KEYNAME and VALUE must be supplied')
        return
    
    apikey = ApiKey(author,application)

    if not plist:
        apikey.updatekey(keyname,keyvalue)
        
    else:
        for keyname,value in apikey.list():
            print('{}={}'.format(keyname, value))


# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    main()




#!/usr/bin/python
###########################################################################################
#   filtercsv - filter a csv file based on indicated filter
#
#   Date        Author      Reason
#   ----        ------      ------
#   05/08/13    Lou King    Create
#
#   Copyright 2013 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
filtercsv - filter a csv file based on indicated filter
===============================================================

Usage::

    TBA
                            
    
'''

# standard
import argparse
import csv
import textwrap
import json
import sys

# pypi

# github

# other

# home grown
from . import version

class invalidParameter(Exception): pass

#----------------------------------------------------------------------
def main():
#----------------------------------------------------------------------
    '''
    filter standard input to standard output
    '''
    descr = textwrap.dedent('''\
            reads stdin and applies FILTER argument to produce stdout
    
            FILTER is a string representing a dict or list of dicts.
            
            The keys of each dict are column headers for which all of the values
            of the dict must be matched for the filter to pass a row in the input
            file (AND function)
            
            If the filter needs to test that a value of the filter and the row NOT
            match, TBD
            
            if a list of dicts is provided, the row passes if any of the dicts match
            (OR function)
            ''')
    parser = argparse.ArgumentParser(prog='filtercsv',
                                     description=descr,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     version='{0} {1}'.format('loutilities',version.__version__))
    parser.add_argument('filter',help='list of dicts or single dict. All items within dict must match for any dict in the list to pass filter')
    args = parser.parse_args()
    
    # convert stdout to binary mode if on windows
    if sys.platform == "win32":
        import os, msvcrt
        msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

    # normalize filter to list of dicts
    #print 'FILTER={}'.format(args.filter)
    filt = json.loads(args.filter)
    if not isinstance(filt, list):
        filt = [filt]
    for f in filt:
        if not isinstance(f, dict):
            raise invalidParameter('FILTER must be dict or list of dicts')
    
    # get the header line, which is always sent to the output file
    hdr = next(sys.stdin)
    sys.stdout.write(hdr)

    # use csv to get a list of the hdr, to handle any errant commas
    H = csv.reader([hdr])
    hdrlist = next(H)
    
    # parse the rest of stdin as a DictReader, stdout as DictWriter
    IN = csv.DictReader(sys.stdin,fieldnames=hdrlist)
    OUT = csv.DictWriter(sys.stdout,hdrlist)
    
    # for each input line, check against all the filters
    for line in IN:
        for f in filt:
            # hope for the best
            match = True
            
            # AND function -- all within dict must match
            for col in list(f.keys()):
                if f[col] != line[col]:
                    match = False
                    break
            
            # OR function -- match found within any of the dicts means pass through the filter
            if match:
                break
        
        # if match found, send to output
        if match:
            OUT.writerow(line)
            
    # clean up
    sys.stdin.close()
    sys.stdout.close()
    
# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
###########################################################################################
#   googleapi - connect to google api through oauth 2
#
#   Date        Author      Reason
#   ----        ------      ------
#   12/02/17    Lou King    adapted from https://developers.google.com/identity/protocols/OAuth2WebServer
#   02/15/19    Lou King    updated due to demise of google+ api
#
#   Copyright 2017 Lou King
###########################################################################################

# standard
import os.path

#pypi
from flask import jsonify, url_for, session, request, abort
from flask.views import View
import httplib2
from oauth2client import client
from google.oauth2 import service_account
from googleapiclient.discovery import build
from googleapiclient.http import MediaFileUpload

# note as of this writing, oauth2client is deprecated.
# See https://github.com/GoogleCloudPlatform/google-auth-library-python/blob/master/docs/oauth2client-deprecation.rst
# but there is no support in the replacement lib for Storage, and google claims they will maintain without adding features
# so this is probably ok
from oauth2client.file import Storage

class parameterError(Exception): pass

class GoogleAuth(View):
    '''
    :param app: flask application
    :param client_secrets_file: client_secrets.json path
    :param scopes: list of google scopes. see https://developers.google.com/identity/protocols/googlescopes
    :param startendpoint: endpoint to start with after authorization completed (no leading slash)
    :param credfolder: folder where credential Storage will be placed
    :param logincallback: function(email) called when login detected
    :param logoutcallback: function called when logout detected
    :param loginfo: info logger function
    :param logdebug: debug logger function
    :param logerror: debug logger function
    '''

    def __init__( self, app, client_secrets_file, scopes, startendpoint, credfolder=None,
                  logincallback=lambda email: None, logoutcallback=lambda email: None,
                  loginfo=None, logdebug=None, logerror=None, ):
        self.app = app
        self.client_secrets_file = client_secrets_file
        self.scopes = scopes
        self.startendpoint = startendpoint
        self.credfolder = credfolder
        self.logincallback = logincallback
        self.logoutcallback = logoutcallback
        self.loginfo = loginfo
        self.logdebug = logdebug
        self.logerror = logerror

        # create supported endpoints
        # _token is ajax call from google sign-in
        self.app.add_url_rule('/_token', view_func=self.token, methods=['POST',])
        self.app.add_url_rule('/clear', view_func=self.clear_credentials, methods=['GET',])

    def token(self):
        # (Receive auth_code by HTTPS POST)
        auth_code = request.data

        # If this request does not have `X-Requested-With` header, this could be a CSRF
        if not request.headers.get('X-Requested-With'):
            abort(403)

        # Exchange auth code for access token, refresh token, and ID token
        credentials = client.credentials_from_clientsecrets_and_code(
            self.client_secrets_file,
            self.scopes,
            auth_code)

        # Get profile info from ID token
        user_id = credentials.id_token['sub']
        email = credentials.id_token['email']

        authorized = False
        if user_id:
            credfile = os.path.join(self.credfolder, user_id)
            storage = Storage(credfile)
            # do the new credentials not have a refresh token? do we already have credentials? 
            # if so, just update the access token and expiry from new credentials into stored and resave
            storedcred = storage.get()
            if not credentials.refresh_token and storedcred:
                storedcred.access_token = credentials.access_token
                storedcred.token_expiry = credentials.token_expiry
                credentials = storedcred
            credentials.set_store(storage)
            storage.put(credentials)
            session['_ga_google_user_id'] = user_id
            session['_ga_google_email'] = email
            # refresh
            http = httplib2.Http()
            credentials.refresh(http)

            # take care of login specifics
            authorized = self.logincallback(email)

        if self.logdebug: self.logdebug( 'oauth2callback() session = {}'.format(session) )

        return jsonify( {'authorized':authorized, 'redirect':url_for(self.startendpoint)} )

    def clear_credentials(self):
        if 'credentials' in session:
            del session['credentials']
        if '_ga_google_user_id' in session:
            del session['_ga_google_user_id']
        
        email = None
        if '_ga_google_email' in session:
            email = session['_ga_google_email']
            del session['_ga_google_email']

        # take care of logout specifics
        self.logoutcallback(email)

        return 'Credentials have been cleared from session cookie'

    def get_userid(self, credentials):
        try:
            user_id = session['_ga_google_user_id']
            if self.logdebug: self.logdebug( 'get_userid() retrieved user_id from session cookie' )

        except KeyError:
            try:
                user_id = credentials.id_token['sub']
                if self.logdebug: self.logdebug( 'get_userid() profile from credentials = {}'.format(credentials) )
            
            except:
                if self.logdebug: self.logdebug( 'invalid credentials, continuing' )
                user_id = None

        return user_id

class GoogleAuthService():
    """
    methods to interact with google api using service based credentials (e.g., for server authentication)

    * see https://developers.google.com/identity/protocols/oauth2/service-account

    :param service_account_file: service_account_file.json path
    :param scopes: list of google scopes. see https://developers.google.com/identity/protocols/googlescopes
    :param loginfo: (optional) info logger function
    :param logdebug: (optional) debug logger function
    :param logerror: (optional) error logger function
    """

    def __init__(self, service_account_file, scopes, loginfo=None, logdebug=None, logerror=None):
        self.service_account_file = service_account_file
        self.scopes = scopes
        self.loginfo = loginfo
        self.logdebug = logdebug
        self.logerror = logerror
        self.credentials = service_account.Credentials.from_service_account_file(self.service_account_file, scopes=self.scopes)
        # see https://stackoverflow.com/a/44518587/799921
        self.drive = build('drive', 'v3', credentials=self.credentials, cache_discovery=False)

    def build(self, googleapi, apiversion):
        """
        build google api service object

        :param googleapi: google api name
        :param apiversion: google api version
        :return: google api service object
        """
        # see https://stackoverflow.com/a/44518587/799921
        return build(googleapi, apiversion, credentials=self.credentials, cache_discovery=False)

    def create_file(self, folderid, filename, contents, doctype='html'):
        """
        create file in drive folder

        ..note::
            folderid must be shared read/write with services account email address

        :param folderid: drive id for folder file needs to reside in
        :param filename: name for file on drive
        :param contents: path for file contents (docx formatted)
        :param doctype: 'html' or 'docx', default 'html'
        :return: google drive id for created file
        """
        ## upload (adapted from https://developers.google.com/drive/api/v3/manage-uploads)
        file_metadata = {
            'name': filename,
            # see https://developers.google.com/drive/api/v3/mime-types
            'mimeType': 'application/vnd.google-apps.document',
            # see https://developers.google.com/drive/api/v3/folder
            'parents': [folderid],
        }

        # mimetype depends on doctype
        mimetype = _getmimetype(doctype)

        # create file
        media = MediaFileUpload(
            contents,
            mimetype=mimetype,
            resumable=True
        )
        file = self.drive.files().create(
            body=file_metadata,
            media_body=media,
            fields='id'
        ).execute()
        fileid = file.get('id')
        return fileid

    def update_file(self, fileid, contents, doctype, new_revision=False, filename=None):
        """
        update file in drive folder

        ..note::
            fileid must be shared read/write with services account email address

        :param fileid: google drive id for file to update
        :param contents: path for file contents (docx formatted)
        :param doctype: 'html' or 'docx'
        :param new_revision: True to create new revision in file, default False
        :param filename: (optional) rename file to this name
        :return: metadata for updated file
        """
        # adapted from https://developers.google.com/drive/api/v2/reference/files/update
        # but see https://stackoverflow.com/questions/35143283/google-drive-api-v3-migration
        file_metadata = self.drive.files().get(fileId=fileid).execute()
        # can't update id
        file_metadata.pop('id')

        # optionally rename file
        if filename:
            file_metadata['name'] = filename

        # mimetype depends on doctype
        mimetype = _getmimetype(doctype)

        # create file
        media = MediaFileUpload(
            contents,
            mimetype=mimetype,
            resumable=True
        )
        file = self.drive.files().update(
            fileId=fileid,
            body=file_metadata,
            media_body=media,
            # newRevision=new_revision,
        ).execute()
        return file

    def set_permission(self, fid, permission):
        # see https://developers.google.com/drive/api/v3/manage-sharing
        def batch_callback(request_id, response, exception):
            if exception:
                # Handle error
                if self.logerror: self.logerror("batch_callback(): permission exception {}".format(exception) )
                raise PermissionError(exception)
            else:
                if self.logdebug: self.logdebug("batch_callback(): permission id {}".format(response.get('id')) )

        batch = self.drive.new_batch_http_request(callback=batch_callback)
        batch.add(self.drive.permissions().create(
                fileId=fid,
                body=permission,
                fields='id',
        ))
        batch.execute()
        
def _getmimetype(doctype):
    """
    determine mimetype for specified doctype
    :param doctype: 'html' or 'docx'
    :return: mimetype
    """
    if doctype not in ['docx', 'html']:
        raise parameterError('_getmimetype(): doctype must be "docx" or "html", found {}'.format(doctype))

    if doctype == 'docx':
        mimetype = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
    else:  # html
        mimetype = 'text/html'

    return mimetype

def get_credentials(credfolder):
    try:
        user_id = session['_ga_google_user_id']
        credfile = os.path.join(credfolder, user_id)
        storage = Storage(credfile)
        credentials = storage.get()
        credentials.set_store(storage)

    except KeyError:
        credentials = None

    return credentials

def get_email():
    try:
        email = session['_ga_google_email']

    except KeyError:
        email = None

    return email




'''
agegrade - calculate age grade statistics
===================================================
'''

# standard
import argparse
import csv
import os.path
import pickle
import shutil

# home grown

# pypi
# github

# exceptions for this module.  See __init__.py for package exceptions
class missingConfiguration(Exception): pass
class parameterError(Exception): pass

def getagtable(agegradewb):
    '''
    in return data structure:
    
    dist is distance in meters (approx)
    openstd is number of seconds for open standard for this distance
    age is age in years (integer)
    factor is age grade factor
    
    :param agegradewb: excel workbook containing age grade factors (e.g., from http://www.howardgrubb.co.uk/athletics/data/wavacalc10.xls)
    
    :rtype: dict: {
            'road: {'F':{dist:{'OC':openstd,age:factor,age:factor,...},...},'M':{dist:{'OC':openstd,age:factor,age:factor,...},...}},
            'trail: {'F':{dist:{'OC':openstd,age:factor,age:factor,...},...},'M':{dist:{'OC':openstd,age:factor,age:factor,...},...}},
            }

    '''
    
    agegradedata = {
        'road': {
            'F': {},
            'M': {},
            'X': {},
        },
        'track': {
            'F': {},
            'M': {},
            'X': {},
        },
    }
    
    # convert the workbook to csv
    from .csvwt import Xls2Csv

    c = Xls2Csv(agegradewb)
    # note non-binary uses men's age grades
    gen2sheet = {'F':'Women', 'M':'Men', 'X':'Men'}
    sheets = c.getfiles()
    
    for gen in ['F', 'M', 'X']:
        SHEET = open(sheets[gen2sheet[gen]],'r',newline='')
        sheet = csv.DictReader(SHEET)
        
        # convert fields to keys - e.g., '5.0' -> 5, skipping non-numeric keys
        f2age = {}
        for f in sheet.fieldnames:
            try:
                k = int(float(f))
                f2age[f] = k
            except ValueError:
                pass
        
        # add each row to data structure, but skip non-running events
        for r in sheet:
            if r['dist(km)'] == '0.0': continue
            
            # dist is rounded to the nearest meter so it can be used as a key
            dist = int(round(float(r['dist(km)'])*1000))

            # create dist
            surface = 'road' if r['isRoad'] == 1 else 'track'
            openstd = float(r['OC'])
            agegradedata[surface][gen][dist] = {'OC':openstd}
            
            # add each age factor
            for f in sheet.fieldnames:
                if f in f2age:
                    age = f2age[f]
                    agegradedata[surface][gen][dist][age] = float(r[f])
            
            
        SHEET.close()
    
    del c
    return agegradedata


class AgeGrade():
    '''
    AgeGrade object 
    
    agegradewb is in format per http://www.howardgrubb.co.uk/athletics/wmalookup15.html (deprecated)
    if agegradewb parameter is missing, previous configuration is used (deprecated)
    configuration is created through command line: agegrade.py [-a agworkbook | -c agconfigfile] (deprecated)
    
    :param agegradedata: data structure used by this class
        {
            'road: {
                'F':{dist:{'OC':openstd,age:factor,age:factor,...},...},
                'M':{dist:{'OC':openstd,age:factor,age:factor,...},...},
                'X':{dist:{'OC':openstd,age:factor,age:factor,...},...},
            },
            'trail: {
                'F':{dist:{'OC':openstd,age:factor,age:factor,...},...},
                'M':{dist:{'OC':openstd,age:factor,age:factor,...},...},
                'X':{dist:{'OC':openstd,age:factor,age:factor,...},...},
            },
        }
    :param agegradewb: (deprecated) excel workbook containing age grade factors
    :param DEBUG: logger function for debug output
    '''
    def __init__(self, agegradedata=None, agegradewb=None, DEBUG=None):
        from .config import CONFIGDIR
        self.DEBUG = DEBUG

        # use age grade data structure if specified
        if agegradedata:
            self.agegradedata = agegradedata
            
        # use age grade workbook if specified
        elif agegradewb:
            self.agegradedata = getagtable(agegradewb)
        
        # otherwise, pick up the data from the configuration
        else:
            pathn = os.path.join(CONFIGDIR,'agegrade.cfg')
            if not os.path.exists(pathn):
                raise missingConfiguration('agegrade configuration not found, run agegrade.py to configure')
            
            C = open(pathn)
            self.agegradedata = pickle.load(C)
            C.close()
            
    def getfactorstd(self, surface, age, gen, distmeters):
        '''
        interpolate factor and openstd based on distance for this age
        
        :param surface: 'road' or 'track'
        :param age: integer age.  If float is supplied, integer portion is used (no interpolation of fractional age)
        :param gen: gender - M or F
        :param distmeters: distance (meters)
        
        :rtype: (factor, openstd) - factor (age grade factor) is between 0 and 1, openstd (open standard) is in seconds
        '''
        
        # round distmeters to the nearest meter as distlist keys are rounded to the nearest meter
        # need to do this so exact match doesn't get interpolated
        # alternate, possibly better, solution would be to use keys in millimeters rather than meters
        # but that would require reloading the data -- that might be fine, but is a bit risky
        # don't use int because need float arithmetic for interpolate
        distmeters = round(distmeters)

        # find surrounding Xi points, and corresponding Fi, OCi points
        distlist = sorted(list(self.agegradedata[surface][gen].keys()))
        lastd = distlist[0]
        for i in range(1,len(distlist)):
            if distmeters <= distlist[i]:
                x0 = lastd
                x1 = distlist[i]
                f0 = self.agegradedata[surface][gen][x0][age]
                f1 = self.agegradedata[surface][gen][x1][age]
                oc0 = self.agegradedata[surface][gen][x0]['OC']
                oc1 = self.agegradedata[surface][gen][x1]['OC']
                break
            lastd = distlist[i]
            
        # interpolate factor and openstd (see http://en.wikipedia.org/wiki/Linear_interpolation)
        factor = f0 + (f1-f0)*((distmeters-x0)/(x1-x0))
        openstd = oc0 + (oc1-oc0)*((distmeters-x0)/(x1-x0))
        
        return factor,openstd
    
    def agegrade(self, age, gen, distmiles, time, surface=None, errorlogger=None):
        '''
        returns age grade statistics for the indicated age, gender, distance, result time
        
        NOTE: non-binary gen X currently returns Men's age grade
        
        :param age: integer age.  If float is supplied, integer portion is used (no interpolation of fractional age)
        :param gen: gender - M, F, X
        :param distmiles: distance (miles)
        :param time: time for distance (seconds)
        :param surface: (optional) 'road' or 'track', default 'road'
        
        :rtype: (age performance percentage, age graded result, age grade factor) - percentage is between 0 and 100, result is in seconds
        '''
        
        # check for some input errors
        gen = gen.upper()
        if gen not in ['F', 'M', 'X']:
            raise parameterError('gen must be M, F, or X')

        # number of meters in a mile
        mpermile = 1609.344     # wavacalc15 has been corrected - now also handles integer distmiles
        
        # some known conversions
        cdist = {26.2:42195,13.1:21098} # wavacalc15 has been corrected
        
        # determine distance in meters
        if distmiles in cdist:
            distmeters = cdist[distmiles]
        else:
            distmeters = distmiles*mpermile
        
        # surface might require some adjustment
        ## if surface not provided, assume road if we have road factors for this distance, else assume track
        ## this maintains backwards compatibility, or for when caller has no access to what surface a race was run on
        initialsurface = surface
        if not surface:
            minroad = min(list(self.agegradedata['road'][gen].keys()))
            # need to round distmeters here because dist keys are rounded integers
            if int(round(distmeters)) >= minroad:
                surface = 'road'
            else:
                surface = 'track'

        ## there are no trail factors, so use road factors if trail requested
        elif surface == 'trail':
            surface = 'road'
        
        # check distance within range.  Make min and max float so exception format specification works
        distlist = list(self.agegradedata[surface][gen].keys())
        minmeters = min(distlist)*1.0
        maxmeters = max(distlist)*1.0
        epsilon = 1 # meter fuzziness
        if distmeters < minmeters-epsilon or distmeters > maxmeters+epsilon:
            if errorlogger:
                errorlogger(f'received age={age} gen={gen} distmiles={distmiles} surface={initialsurface} time={time}, used surface={surface}')
            raise parameterError(
                'distmiles must be between {:0.3f} and {:0.1f}'.format(minmeters / mpermile, maxmeters / mpermile))

        # interpolate factor and openstd based on distance for this age
        age = int(age)
        if age in range(5,100):
            factor,openstd = self.getfactorstd(surface, age, gen, distmeters)
        
        # extrapolate for ages < 5
        elif age < 5:
            if True:
                factor,openstd = self.getfactorstd(surface, 5, gen, distmeters)
            
            # don't do extrapolation
            else:
                age1 = 5
                age2 = 6
                factor1,openstd1 = self.getfactorstd(age1,gen,distmeters)
                factor2,openstd2 = self.getfactorstd(age2,gen,distmeters)
                factor = factor1 + (1.0*(age-age1)/(age2-age1))*(factor2-factor1)
                openstd = openstd1 + (1.0*(age-age1)/(age2-age1))*(openstd2-openstd1)
            
         # extrapolate for ages > 99
        elif age > 99:
            if True:
                factor,openstd = self.getfactorstd(surface, 99, gen, distmeters)
            
            # don't do extrapolation
            else:    
                age1 = 98
                age2 = 99
                factor1,openstd1 = self.getfactorstd(age1,gen,distmeters)
                factor2,openstd2 = self.getfactorstd(age2,gen,distmeters)
                factor = factor1 + (1.0*(age-age1)/(age2-age1))*(factor2-factor1)
                openstd = openstd1 + (1.0*(age-age1)/(age2-age1))*(openstd2-openstd1)
        
        # return age grade statistics
        agpercentage = 100*(openstd/factor)/time
        agresult = time*factor
        if self.DEBUG:
            self.DEBUG(f'dist={distmeters} surf={surface} age={age} gen={gen} openstd={openstd} factor={factor} time={time} agresult={agresult} ag%={agpercentage}\n')
        return agpercentage, agresult, factor

    def result(self,surface, age, gen, distmiles, agpc):
        '''
        returns age grade statistics for the indicated age, gender, distance, result time

        NOTE: non-binary gen X currently returns Men's age grade

        :param surface: 'road' or 'track'
        :param age: integer age.  If float is supplied, integer portion is used (no interpolation of fractional age)
        :param gen: gender - M, F, X
        :param distmiles: distance (miles)
        :param agpc: age grade percentage - between 0 and 100

        :rtype: result in seconds
        '''

        # check for some input errors
        gen = gen.upper()
        if gen not in ['F', 'M', 'X']:
            raise parameterError('gen must be M, F, or X')

        # number of meters in a mile -- close enough for this data set
        mpermile = 1609.344

        # some known conversions
        cdist = {26.2:42195,13.1:21098}

        # determine distance in meters
        if distmiles in cdist:
            distmeters = cdist[distmiles]
        else:
            distmeters = distmiles*mpermile

        # check distance within range.  Make min and max float so exception format specification works
        distlist = list(self.agegradedata[surface][gen].keys())
        minmeters = min(distlist)*1.0
        maxmeters = max(distlist)*1.0
        if distmeters < minmeters or distmeters > maxmeters:
            raise parameterError('distmiles must be between {0:f0.3} and {1:f0.1}'.format(minmeters/mpermile,maxmeters/mpermile))

        # interpolate factor and openstd based on distance for this age
        age = int(age)
        if age in range(5,100):
            factor,openstd = self.getfactorstd(age,gen,distmeters)

        # extrapolate for ages < 5
        elif age < 5:
            # don't do extrapolation
            if True:
                factor,openstd = self.getfactorstd(5,gen,distmeters)

            else:
                age1 = 5
                age2 = 6
                factor1,openstd1 = self.getfactorstd(age1,gen,distmeters)
                factor2,openstd2 = self.getfactorstd(age2,gen,distmeters)
                factor = factor1 + (1.0*(age-age1)/(age2-age1))*(factor2-factor1)
                openstd = openstd1 + (1.0*(age-age1)/(age2-age1))*(openstd2-openstd1)

         # extrapolate for ages > 99
        elif age > 99:
            if True:
                factor,openstd = self.getfactorstd(99,gen,distmeters)

            # don't do extrapolation
            else:
                age1 = 98
                age2 = 99
                factor1,openstd1 = self.getfactorstd(age1,gen,distmeters)
                factor2,openstd2 = self.getfactorstd(age2,gen,distmeters)
                factor = factor1 + (1.0*(age-age1)/(age2-age1))*(factor2-factor1)
                openstd = openstd1 + (1.0*(age-age1)/(age2-age1))*(openstd2-openstd1)

        # return result
        time = (openstd/factor)/(agpc/100.0)
        return time

def main(): 
    descr = '''
    Update configuration for agegrade.py.  One of --agworkbook or --agconfigfile must be used,
    but not both.
    
    --agworkbook creates an agconfigfile and puts it in the configuration directory.
    --agconfigfile simply places the indicated file into the configuration directory.
    '''

    from . import version
    from .config import CONFIGDIR

    parser = argparse.ArgumentParser(prog='loutilities', description=descr,formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {}'.format(version.__version__))
    parser.add_argument('-a','--agworkbook',help='filename of age grade workbook.', default=None)
    parser.add_argument('-c','--agconfigfile',help='filename of age grade config file',default=None)
    args = parser.parse_args()

    # must have one of the options
    if not args.agworkbook and not args.agconfigfile:
        print('one of --agworkbook or --agconfigfile must be specified')
        return
        
    # can't have both of the options
    if args.agworkbook and args.agconfigfile:
        print('only one of --agworkbook or --agconfigfile should be specified')
        return

    # configuration file will be here    
    pathn = os.path.join(CONFIGDIR,'agegrade.cfg')

    # workbook specified
    if args.agworkbook:
        agegradedata = getagtable(args.agworkbook)
        C = open(pathn,'w')
        pickle.dump(agegradedata,C)
        C.close()
    
    # config file specified
    else:
        # make sure this is a pickle file
        try:
            C = open(args.agconfigfile)
            test = pickle.load(C)
            C.close()
        except IOError:
            print('{0}: not found'.format(args.agconfigfile))
            return
        except KeyError:
            print('{0}: invalid format'.format(args.agconfigfile))
            return
            
        shutil.copyfile(args.agconfigfile,pathn)

    print('updated {0}'.format(pathn))


# ##########################################################################################
#	__main__
# ##########################################################################################
if __name__ == "__main__":
    main()


#!/usr/bin/python

# Copyright 2012 Lou King
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import pdb

# home grown
from loutilities import version

from setuptools import setup, find_packages

setup(
    name = "loutilities",
    version = version.__version__,
    packages = find_packages(),
#    include_package_data = True,
    scripts = [
        'loutilities/agegrade.py',
        'loutilities/apikey.py',
        'loutilities/applytemplate.py',
        'loutilities/filtercsv.py',
        'loutilities/makerst.py',
    ],

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires = [
          # 'unicodecsv>=0.13.0',
          'flask-security-too>=5.0.0',
          'flask-sqlalchemy>=3.0.5',
        ],

    # include data files as appropriate
    package_data={
        '': [
            'tables-assets/static/*',
            'tables-assets/static/user/*',
            'tables-assets/static/user/admin/*',
            'tables-assets/templates/*',
            'tables-assets/templates/security/*',
            'tables-assets/templates/security/email/*',
            '*.csv',
        ],
    },

    entry_points = {
        'console_scripts': [
            'agegrade = loutilities.agegrade:main',
            'apikey = loutilities.apikey:main',
            'applytemplate = loutilities.applytemplate:main',
            'filtercsv = loutilities.filtercsv:main',
            'makerst = loutilities.makerst:main',
        ],
    },


    zip_safe = False,

    # metadata for upload to PyPI
    description = 'some hopefully useful utilities',
    long_description=open("README.md").read(),
    license = 'Apache License, Version 2.0',
    author = 'Lou King',
    author_email = 'lking@pobox.com',
    url = 'http://github.com/louking/loutilities',
    # could also include long_description, download_url, classifiers, etc.
)




'''
views - flask view support
'''
# standard
from urllib.parse import urlencode

# pypi
from flask import request, redirect, url_for, g, flash

# homegrown
from . import DbCrudApiInterestsRolePermissions

class NotImplementedError(Exception): pass

class SelectInterestsView(DbCrudApiInterestsRolePermissions):
    '''
    create a view with a single select, and a submit button, based on DbCrudApiInterestsRolePermissions for the
    user permissions stuff, otherwise gutted

    :param app: app or blueprint instance
    :param template: (optional) template file to use (defaults to select-view.jinja2)
    :param pagename: page name for title on page
            can also be a function which returns this parameter
    :param endpoint: endpoint parameter used by flask.url_for()
    :param rule: rule parameter used by flask.add_url_rule() [defaults to '/' + endpoint]
    :param displayonly: set to True if the form should only be displayed (no submit option), default False
            can also be a function which returns this parameter
    :param preselecthtml: string any html which needs to go before the select
            can also be a function which returns this parameter
    :param select2options: options for select2 initialization (https://select2.org/configuration/options-api)
            can also be a function which returns this parameter
    '''
    def __init__(self, **kwargs):
        # the args dict has all the defined parameters to
        # caller supplied keyword args are used to update the defaults
        # all arguments are made into attributes for self
        self.kwargs = kwargs
        args = dict(template = 'select-view.jinja2',
                    # pagename = None,
                    # endpoint = None,
                    # rule = None,
                    preselecthtml = '',
                    select2options = {},
                    selectlabel = None,
                    posturl=lambda: self._pageurl(),
                    displayonly=False,
                    )
        args.update(kwargs)

        # expected arguments for DbCrudApiInterestsRolePermissions
        args.update(
            {
                'clientcolumns': [],
            }
        )
        super().__init__(**args)

    def _pageurl(self):
        baseurl = url_for(self.endpoint, interest=g.interest)
        params = urlencode(request.args)
        return '{}?{}'.format(baseurl, params)

    def register(self):
        # name for view is last bit of fully named endpoint
        name = self.endpoint.split('.')[-1]

        # create supported endpoint
        self.my_view = self.as_view(name, **self.kwargs)
        self.app.add_url_rule('{}'.format(self.rule),view_func=self.my_view,methods=['GET', 'POST'])

    # def render_template(self, **kwargs):
    #     return render_template(self.template, **kwargs)

    def getval(self):
        '''
        must be replaced
        :return: value to use for initial val of select
        '''
        raise NotImplementedError

    def putval(self, val):
        '''
        must be replaced
        :param val: value to set based on val of select
        '''
        raise NotImplementedError

    def _template_options(self):
        return dict(
            select2options=self.select2options if not callable(self.select2options) else self.select2options(),
            select2val=self.getval(),
            pagename=self.pagename if not callable(self.pagename) else self.pagename(),
            preselecthtml=self.preselecthtml if not callable(self.preselecthtml) else self.preselecthtml(),
            selectlabel=self.selectlabel if not callable(self.selectlabel) else self.selectlabel(),
            posturl=self.posturl if not callable(self.posturl) else self.posturl(),
            displayonly=self.displayonly if not callable(self.displayonly) else self.displayonly(),
        )

    def get(self):
        if not self.permission():
            self.abort()

        # render page
        return self.render_template(**self._template_options())

    def post(self):
        if not self.permission():
            self.abort()

        val = request.form.get('select-select', None)
        self.putval(val)
        flash('Submission successful')
        return redirect(self._pageurl())




###########################################################################################
# settings - define default, test and production settings
#
#       Date            Author          Reason
#       ----            ------          ------
#       03/06/20        Lou King        Create
#
#   Copyright 2020 Lou King.  All rights reserved
#
###########################################################################################
'''
settings - define default, test and production settings

see http://flask.pocoo.org/docs/1.0/config/?highlight=production#configuration-best-practices
'''

# standard
import logging

# homegrown
from loutilities.configparser import getitems


class Config(object):
    DEBUG = False
    TESTING = False

    # default database
    # https://flask-sqlalchemy.palletsprojects.com/en/2.x/binds/
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    SQLALCHEMY_BINDS = {
        'users': 'sqlite:///:memory:',
    }

    # logging
    LOGGING_LEVEL_FILE = logging.INFO
    LOGGING_LEVEL_MAIL = logging.ERROR

    # flask-security configuration -- see https://pythonhosted.org/Flask-Security/configuration.html
    SECURITY_TRACKABLE = True
    SECURITY_DEFAULT_REMEMBER_ME = True

    # avoid warning
    SQLALCHEMY_TRACK_MODIFICATIONS = False

class Testing(Config):
    TESTING = True
    WTF_CSRF_ENABLED = False

    # need to set SERVER_NAME to something, else get a RuntimeError about not able to create URL adapter
    # must have following line in /etc/hosts or C:\Windows\System32\drivers\etc\hosts file
    #   127.0.0.1 dev.localhost
    SERVER_NAME = 'dev.localhost'

    # need a default secret key - in production replace by config file
    SECRET_KEY = "<test secret key>"

    # need to allow logins in flask-security. see https://github.com/mattupstate/flask-security/issues/259
    LOGIN_DISABLED = False


class RealDb(Config):
    def __init__(self, configfiles):
        if type(configfiles) == str:
            configfiles = [configfiles]

        # connect to database based on configuration
        config = {}
        for configfile in configfiles:
            config.update(getitems(configfile, 'database'))

        # https://flask-sqlalchemy.palletsprojects.com/en/2.x/binds/
        userdbuser = config['userdbuser']
        userpassword = config['userdbpassword']
        userdbserver = config['userdbserver']
        userdbname = config['userdbname']
        userdb_uri = 'mysql://{uname}:{pw}@{server}/{dbname}'.format(uname=userdbuser, pw=userpassword, server=userdbserver,
                                                                 dbname=userdbname)
        self.SQLALCHEMY_BINDS = {
            'users': userdb_uri
        }


class Development(RealDb):
    DEBUG = True


class Production(RealDb):
    pass





###########################################################################################
#       Date            Author          Reason
#       ----            ------          ------
#       07/03/18        Lou King        Create
#
#   Copyright 2018 Lou King.  All rights reserved
###########################################################################################
'''
models - dummy unit test models
=====================================================
from https://raw.githubusercontent.com/louking/sqlalchemy-datatables/master/tests/models.py

'''

import datetime

from sqlalchemy import Column, Integer, String, Boolean, DateTime, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.hybrid import hybrid_property

Base = declarative_base()


class User(Base):

    """Define a User."""

    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    created_at = Column(DateTime, default=datetime.datetime.utcnow)
    address = relationship('Address', uselist=False, backref=backref('user'))

    def __unicode__(self):
        """Give a readable representation of an instance."""
        return '%s' % self.name

    def __repr__(self):
        """Give a unambiguous representation of an instance."""
        return '<%s#%s>' % (self.__class__.__name__, self.id)

    @hybrid_property
    def dummy(self):
        return '%s%s-DUMMY' % (self.name[0:1], str(self.id))

    @dummy.expression
    def dummy(cls):
        return True


class Address(Base):

    """Define an Address."""

    __tablename__ = 'addresses'

    id = Column(Integer, primary_key=True)
    description = Column(String, unique=True)
    user_id = Column(Integer, ForeignKey('users.id'))

    def __unicode__(self):
        """Give a readable representation of an instance."""
        return '%s' % (self.id)

    def __repr__(self):
        """Give a unambiguous representation of an instance."""
        return '<%s#%s>' % (self.__class__.__name__, self.id)

class NotUnique(Base):
    '''
    table with values which are not necessarily unique
    '''
    __tablename__ = 'notunique'

    id = Column(Integer, primary_key=True)
    value = Column(String)

class SeveralAttrs(Base):
    '''
    table with several attributes
    '''
    __tablename__ = 'severalattrs'
    id = Column(Integer, primary_key=True)
    intAttr1  = Column( Integer )
    strAttr2  = Column( String )
    strAttr3  = Column( String )
    boolAttr4 = Column( Boolean )
    dateAttr5 = Column( DateTime )

